import io
import asyncio
import threading
import queue
import logging

from starlette import concurrency


logger = logging.getLogger('vaex.file.async')


class WriteStream(io.RawIOBase):
    '''File like object that has a sync write API, and a generator as consumer.
    
    This is useful for letting 1 thread write to this object, while another thread
    in consuming the byte chunks via an iterator, or async iterator.


    To let the writer thread stop, simply close the file.
    '''
    def __init__(self, queue_size=3):
        self.writes = []
        self.queue = queue.Queue(queue_size)
        self.closed = False
        self.pos = 0
        self.exception = None

    def __iter__(self):
        yield from self.chunks()

    async def __aiter__(self):    
        import starlette.concurrency
        async for item in starlette.concurrency.iterate_in_threadpool(self.chunks()):
            yield item

    def __enter__(self, *args):
        return self

    def __exit__(self, type, value, traceback):
        self.close()
        self.exception = value


    def read(self, *args):
        return b''

    def seek(self, *args):
        if self.closed:
            return ValueError('stream closed')
        return 0

    def readinto(self, *args):
        return b''

    def tell(self):
        if self.closed:
            return ValueError('stream closed')
        return self.pos

    def flush(self):
        if self.closed:
            return ValueError('stream closed')
        pass

    def writable(self):
        return True

    def readable(self):
        return False

    def seekable(self):
        return False
    
    def write(self, b):
        if self.closed:
            return ValueError('stream closed')
        logger.debug('write: %r', len(b))
        buffer = memoryview(b)
        # print("--", bytes(buffer))
        self.pos += len(buffer)
        # we need to copy it seems (otherwise if we use a buffered writer, it will reuse the memory)
        self.queue.put(bytes(buffer))
        # print(len(buffer))
        return len(buffer)

    def close(self, force=False):
        '''Note that close can block due to the queue, using force=True the queue will be cleared'''
        logger.debug('closing stream, putting None element in queue to stop chunk yielding')
        # make sure nobody will add new items to the queue, at max 1 (the current write)
        self.closed = True
        if force:
            self._force_put(None)
        else:
            self.queue.put(None)


    def closed(self):
        return self.closed

    def chunks(self):
        logger.debug('yielding chunks')
        while True:
            logger.debug('waiting for chunk')
            item = self.queue.get()
            if item is None:
                logger.debug('stop yielding, file closed')
                break
            # if isinstance(item, BaseException):
            #     logger.debug('stop yielding, exception occured')
            #     raise item
            yield item
        if self.exception:
            raise self.exception

    def getvalue(self):
        return b''.join(bytes(k) for k in self.chunks())

    # def stop(self, exception):
    #     '''Empty the queue, and put the exception on the queue, and close the file.

    #     If we don't empty the queue, this might block
    #     '''
    #     self._force_put(exception)

    #     self.closed = True

    def _force_put(self, item):
        # keep trying to put an exception on the queue
        # we dont want to do this blocking, since a producer
        # thread might fill this up.
        done = False
        while not done:
            logger.debug('clearing queue')
            while not self.queue.empty():
                self.queue.get(block=False)
            try:
                logger.debug('put %r on queue', item)
                self.queue.put(item, block=False)
                done = True
            except queue.Full:
                logger.debug('retry putting exception on queue, because it was filled')



# -*- coding: utf-8 -*-
from __future__ import absolute_import
import ast
import collections
import concurrent.futures
import contextlib
import functools
import json
import math
import os
import platform
import re
import sys
import threading
import time
from typing import MutableMapping
import warnings
import numbers
import keyword
from filelock import FileLock
import pkg_resources

import blake3
import dask.utils
import numpy as np
import pyarrow as pa
import six
import yaml

from .json import VaexJsonEncoder, VaexJsonDecoder
import vaex.file


is_frozen = getattr(sys, 'frozen', False)
PY2 = sys.version_info[0] == 2
PY3 = sys.version_info[0] == 3

osname = dict(darwin="osx", linux="linux", windows="windows")[platform.system().lower()]
# $ export VAEX_DEV=1 to enabled dev mode (skips slow tests)
devmode = os.environ.get('VAEX_DEV', '0') == '1'


# so that vaex can be imported without compiling the extenstions (like in readthedocs)
has_c_extension = os.environ.get('VAEX_NO_C_EXTENSIONS', None) != "1"

class AttrDict(dict):
    def __init__(self, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)

    def __getattr__(self, name):
        return self[name]

    def __setattr__(self, key, value):
        self[key] = value


class RegistryCallable(MutableMapping):
    '''lazily load entries from entry_points'''
    def __init__(self, entry_points : str, typename : str):
        self.entry_points = entry_points
        self.typename = typename
        self.registry = {}
        self.lock = threading.Lock()

    def __getitem__(self, name):
        if name in self.registry:
            return self.registry[name]
        with self.lock:
            if name in self.registry:
                return self.registry[name]
            for entry in pkg_resources.iter_entry_points(group=self.entry_points):
                self.registry[entry.name] = entry.load()

        if name not in self.registry:
            raise NameError(f'No {self.typename} registered with name {name!r} under entry_point {self.entry_points!r}')
        return self.registry[name]

    def __delitem__(self, name):
        del self.registry[name]

    def __iter__(self):
        return self.registry.__iter__()

    def __len__(self):
        return self.registry.__len__()

    def __setitem__(self, name, value):
        self.registry.__setitem__(name, value)


def deprecated(reason):
    def wraps(f):
        if not f.__doc__:
            f.__doc__ = ""
        f.__doc__ = "Deprecated: {}\n\n{}".format(reason, f.__doc__)
        @functools.wraps(f)
        def wraps2(*args, **kwargs):
            warnings.warn("Call to deprecated function {}: {}".format(f.__name__, reason),
                          category=DeprecationWarning, stacklevel=2)
            return f(*args, **kwargs)

        return wraps2

    return wraps


def subdivide(length, parts=None, max_length=None):
    """Yields index tuple (i1, i2) that subdivide an array into parts of max_length"""
    if max_length is None:
        max_length = (length + parts - 1) / parts
    i1 = 0
    while i1 < length:
        i2 = min(i1 + max_length, length)
        yield i1, i2
        i1 = i2
    assert i1 == length


def subdivide_mask(mask, parts=None, max_length=None, logical_length=None):
    """Yields index tuple (l1, l2, i1, i2) that subdivide an array into parts such that it contains max_length True values in mask

    l1 an l2 refer to the logical indices, similar as :func:`subdivide`, while i1, and i2 refer to the 'raw' indices, such that
    `np.sum(mask[i1:i2]) == logical_length` (except for the last element).
    """
    if logical_length is None:
        logical_length = np.asarray(mask).sum()
    if logical_length == 0:
        return
    raw_length = len(np.asarray(mask))
    if max_length is None:
        max_length = (logical_length + parts - 1) / parts
    full_mask = mask
    logical_start = 0
    logical_end =  min(logical_start + max_length, logical_length)
    raw_index = full_mask.raw_offset(1)
    assert raw_index != -1

    while logical_start < logical_length:
        # slice the mask from our offset till end
        mask = full_mask.view(raw_index, raw_length)
        # count how many raw elements we need to skip to get a logical chunk_size
        raw_offset = mask.raw_offset(logical_end - logical_start)
        assert raw_offset != -1
        next_raw_index = raw_index + raw_offset + 1
        yield logical_start, logical_end, raw_index, next_raw_index
        raw_index = next_raw_index
        logical_start = min(logical_start + max_length, logical_length)
        logical_end = min(logical_start + max_length, logical_length)


def submit_subdivide(thread_count, f, length, max_length):
    futures = []
    thread_pool = concurrent.futures.ThreadPoolExecutor(thread_count)
    # thread_pool = concurrent.futures.ProcessPoolExecutor(thread_count)
    for i1, i2 in list(subdivide(length, max_length=max_length)):
        futures.append(thread_pool.submit(f, i1, i2))
    return futures


def linspace_centers(start, stop, N):
    return np.arange(N) / (N + 0.) * (stop - start) + float(stop - start) / N / 2 + start


def multisum(a, axes):
    correction = 0
    for axis in axes:
        a = np.nansum(a, axis=axis - correction)
        correction += 1
    return a


def disjoined(data):
    # create marginalized distributions and multiple them together
    data_disjoined = None
    dim = len(data.shape)
    for d in range(dim):
        axes = list(range(dim))
        axes.remove(d)
        data1d = multisum(data, axes)
        shape = [1 for k in range(dim)]
        shape[d] = len(data1d)
        data1d = data1d.reshape(tuple(shape))
        if d == 0:
            data_disjoined = data1d
        else:
            data_disjoined = data_disjoined * data1d
    return data_disjoined


def get_root_path():
    osname = platform.system().lower()
    # if (osname == "linux") and is_frozen: # we are using pyinstaller
    if is_frozen:  # we are using pyinstaller or py2app
        return os.path.dirname(sys.argv[0])
    else:
        return os.path.abspath(".")


def os_open(document):
    """Open document by the default handler of the OS, could be a url opened by a browser, a text file by an editor etc"""
    osname = platform.system().lower()
    if osname == "darwin":
        os.system("open \"" + document + "\"")
    if osname == "linux":
        cmd = "xdg-open \"" + document + "\"&"
        os.system(cmd)
    if osname == "windows":
        os.system("start \"" + document + "\"")


def filesize_format(value):
    for unit in ['bytes', 'KiB', 'MiB', 'GiB']:
        if value < 1024.0:
            return "%3.1f%s" % (value, unit)
        value /= 1024.0
    return "%3.1f%s" % (value, 'TiB')


log_timer = True


class Timer(object):
    def __init__(self, name=None, logger=None):
        self.name = name
        self.logger = logger

    def __enter__(self):
        if log_timer:
            if self.logger:
                self.logger.debug("%s starting" % self.name)
            else:
                print(('[%s starting]...' % self.name))
            self.tstart = time.time()

    def __exit__(self, type, value, traceback):
        if log_timer:
            msg = "%s done, %ss elapsed" % (self.name, time.time() - self.tstart)
            if self.logger:
                self.logger.debug(msg)
            else:
                print(msg)
            if type or value or traceback:
                print((type, value, traceback))
        return False


def get_vaex_home():
    '''Get vaex home directory, defaults to $HOME/.vaex.

    The $VAEX_HOME environment variable can be set to override this default.

    If both $VAEX_HOME and $HOME are not define, the current working directory is used.
    '''
    if 'VAEX_HOME' in os.environ:
        return os.environ['VAEX_HOME']
    if 'VAEX_PATH_HOME' in os.environ:  # for backwards compatibility
        return os.environ['VAEX_PATH_HOME']
    elif 'HOME' in os.environ:
        return os.path.join(os.environ['HOME'], ".vaex")
    else:
        return os.getcwd()


def get_private_dir(subdir=None, *extra):
    path = get_vaex_home()
    if subdir:
        path = os.path.join(path, subdir, *extra)
    os.makedirs(path, exist_ok=True)
    return path


def make_list(sequence):
    if isinstance(sequence, np.ndarray):
        return sequence.tolist()
    else:
        return list(sequence)


def confirm_on_console(topic, msg):
    done = False
    print(topic)
    while not done:
        output = raw_input(msg + ":[y/n]")
        if output.lower() == "y":
            return True
        if output.lower() == "n":
            return False


def yaml_dump(f, data):
    yaml.safe_dump(data, f, default_flow_style=False, encoding='utf-8', allow_unicode=True, sort_keys=False)


def yaml_load(f):
    return yaml.safe_load(f)


def write_json_or_yaml(file, data, fs_options={}, fs=None, old_style=True):
    file, path = vaex.file.file_and_path(file, mode='w', fs_options=fs_options, fs=fs)
    try:
        if path:
            base, ext = os.path.splitext(path)
        else:
            ext = '.json'  # default
        if ext == ".json":
            json.dump(data, file, indent=2, cls=VaexJsonEncoder if old_style else None)
        elif ext == ".yaml":
            yaml_dump(file, data)
        else:
            raise ValueError("file should end in .json or .yaml (not %s)" % ext)
    finally:
        file.close()


def read_json_or_yaml(file, fs_options={}, fs=None, old_style=True):
    file, path = vaex.file.file_and_path(file, fs_options=fs_options, fs=fs)
    try:
        if path:
            base, ext = os.path.splitext(path)
        else:
            ext = '.json'  # default
        if ext == ".json":
            return json.load(file, cls=VaexJsonDecoder if old_style else None) or {}
        elif ext == ".yaml":
            return yaml_load(file) or {}
        else:
            raise ValueError("file should end in .json or .yaml (not %s)" % ext)
    finally:
        file.close()


def check_memory_usage(bytes_needed, confirm):
    psutil = optional_import('psutil')
    if bytes_needed > psutil.virtual_memory().available:
        if bytes_needed < (psutil.virtual_memory().available + psutil.swap_memory().free):
            text = "Action requires %s, you have enough swap memory available but it will make your computer slower, do you want to continue?" % (
            filesize_format(bytes_needed),)
            return confirm("Memory usage issue", text)
        else:
            text = "Action requires %s, you do not have enough swap memory available, do you want try anyway?" % (
            filesize_format(bytes_needed),)
            return confirm("Memory usage issue", text)
    return True


def ensure_string(string_or_bytes, encoding="utf-8", cast=False):
    if cast:
        if six.PY2:
            string_or_bytes = unicode(string_or_bytes)
        else:
            string_or_bytes = str(string_or_bytes)
    if isinstance(string_or_bytes, six.string_types):
        return string_or_bytes
    else:
        return string_or_bytes.decode(encoding)


def filename_shorten(path, max_length=150):
    # parts = path.split(os.path.sep)
    parts = []
    done = False
    tail = path
    while not done:
        tail, head = os.path.split(tail)
        # print ">>", tail, head
        if not head:
            done = True
            parts.append(tail)
        else:
            parts.append(head)
    parts.reverse()
    # print "parts", parts, os.path.join(*parts)
    if len(parts) > 4:
        first, middle, last = os.path.join(parts[0], parts[1]), parts[2:-1], parts[-1]
        # print first, middle, last
        while (len(os.path.join(first, *(middle + [last]))) <= max_length) and middle:
            last = os.path.join(middle[-1], last)
            middle = middle[:-1]
        if middle:
            return os.path.join(first, "...", last)
        else:
            return os.path.join(first, last)
    else:
        return path


def listify(*args):
    import vaex.expression
    if isinstance(args[0], (six.string_types, vaex.expression.Expression)):
        return False, [[x] for x in args]
    try:
        _ = args[0][0]
        return True, args
    except:
        return False, [[x] for x in args]


def unlistify(waslist, *args):
    if waslist:
        if len(args) == 1:
            return args[0]
        return args
    else:
        values = [x[0] for x in args]
        if len(values) == 1:
            return values[0]


def valid_expression(names, name):
    if name in names and not valid_identifier(name):
        return f'df[%r]' % name
    else:
        return name


def valid_identifier(name):
    return name.isidentifier() and not keyword.iskeyword(name)


def find_valid_name(name, used=None):
    if used is None:
        used = []
    if isinstance(name, int):
        name = str(name)
    if name in used:
        nr = 1
        while name + ("_%d" % nr) in used:
            nr += 1
        name = name + ("_%d" % nr)
    return name


def _python_save_name(name, used=None):
    if used is None:
        used = []
    first, rest = name[0], name[1:]
    name = re.sub("[^a-zA-Z_]", "_", first) + re.sub("[^a-zA-Z_0-9]", "_", rest)
    if name in used:
        nr = 1
        while name + ("_%d" % nr) in used:
            nr += 1
        name = name + ("_%d" % nr)
    return name


@contextlib.contextmanager
def write_to(f, mode):
    """Flexible writing, where f can be a filename or f object, if filename, closed after writing"""
    if hasattr(f, 'write'):
        yield f
    else:
        f = open(f, mode)
        yield f
        f.close()


# these are used to add namespace to vaex' dataset
# like ds.viz.plot2d(ds.x, ds.y)

class BoundMethods(object):

    def __init__(self, obj, methods):
        for name, value in methods.items():
            setattr(self, name, value.__get__(obj))


class InnerNamespace(object):

    def __init__(self, methods, cls=None, prefix=''):
        self._methods = methods
        self.cls = cls
        self.prefix = prefix

    def _add(self, **kwargs):
        self._methods.update(kwargs)
        self.__dict__.update(kwargs)
        if self.cls:
            for name, value in kwargs.items():
                setattr(self.cls, self.prefix + name, value)

    # def __getattr__(self, name):
    # 	if name in self._methods:
    # 		return self._methods[name]
    # 	else:
    # 		return object.__getattr__(self, name)

    def __get__(self, obj, objtype):
        # print("get", obj, objtype)
        if obj is None:
            return self
        else:
            return BoundMethods(obj, self._methods)


def _parse_f(f):
    if f is None:
        return lambda x: x
    elif isinstance(f, six.string_types):
        if f == "identity":
            return lambda x: x
        else:
            if hasattr(np, f):
                return getattr(np, f)
            else:
                raise ValueError("do not understand f = %s, should be a function, string 'identity' or a function from numpy such as 'log', 'log1p'" % f)
    else:
        return f


def _normalize(a, axis=None):
    a = np.copy(a)  # we're gonna modify inplace, better copy iy
    mask = np.isfinite(a)
    a[~mask] = np.nan  # put inf to nan
    allaxis = list(range(len(a.shape)))
    if axis is not None:
        if type(axis) == int:
            axis = [axis]
        for ax in axis:
            allaxis.remove(ax)
        axis = tuple(allaxis)
    vmin = np.nanmin(a)
    vmax = np.nanmax(a)
    a = a - np.nanmin(a, axis=axis, keepdims=True)
    a /= np.nanmax(a, axis=axis, keepdims=True)
    return a, vmin, vmax


def _normalize_selection_name(name):
    if name is True:
        return "default"
    elif name is False:
        return None
    else:
        return name


def _normalize_selection(selection):
    if isinstance(selection, (list, tuple)):
        return type(selection)([_normalize_selection_name(k) for k in selection])
    else:
        return _normalize_selection_name(selection)


def _parse_n(n):
    if isinstance(n, six.string_types):
        if n == "normalize":
            return _normalize
            # return lambda x: x
        else:
            raise ValueError("do not understand n = %s, should be a function, or string 'normalize'" % n)
    else:
        return n


def _parse_reduction(name, colormap, colors):
    if name.startswith("stack.fade"):
        def _reduce_stack_fade(grid):
            return grid[..., -1]  # return last..
        return _reduce_stack_fade
    elif name.startswith("colormap"):
        import matplotlib.cm
        cmap = matplotlib.cm.get_cmap(colormap)

        def f(grid):
            masked_grid = np.ma.masked_invalid(grid)  # convert inf/nan to a mask so that mpl colors bad values correcty
            return cmap(masked_grid)
        return f
    elif name.startswith("stack.color"):
        def f(grid, colors=colors, colormap=colormap):
            import matplotlib.cm
            colormap = matplotlib.cm.get_cmap(colormap)
            if isinstance(colors, six.string_types):
                colors = matplotlib.cm.get_cmap(colors)
            if isinstance(colors, matplotlib.colors.Colormap):
                group_count = grid.shape[-1]
                colors = [colors(k / float(group_count - 1.)) for k in range(group_count)]
            else:
                colors = [matplotlib.colors.colorConverter.to_rgba(k) for k in colors]
            # print grid.shape
            total = np.nansum(grid, axis=0) / grid.shape[0]
            # grid /= total
            # mask = total > 0
            # alpha = total - total[mask].min()
            # alpha[~mask] = 0
            # alpha = total / alpha.max()
            # print np.nanmax(total), np.nanmax(grid)
            return colormap(total)
            rgba = grid.dot(colors)
            # def _norm(data):
            #   mask = np.isfinite(data)
            #   data = data - data[mask].min()
            #   data /= data[mask].max()
            #   return data
            # rgba[...,3] = (f(alpha))
            # rgba[...,3] = 1
            rgba[total == 0, 3] = 0.
            # mask = alpha > 0
            # if 1:
            #   for i in range(3):
            #       rgba[...,i] /= total
            #       #rgba[...,i] /= rgba[...,0:3].max()
            #       rgba[~mask,i] = background_color[i]
            # rgba = (np.swapaxes(rgba, 0, 1))
            return rgba
        return f

    else:
        raise ValueError("do not understand reduction = %s, should be a ..." % name)


def _is_string(x):
    return isinstance(x, six.string_types)


def _issequence(x):
    return isinstance(x, (tuple, list, np.ndarray))


def _isnumber(x):
    return isinstance(x, (numbers.Number, pa.Scalar))


def _is_limit(x):
    return isinstance(x, (tuple, list, np.ndarray)) and all([_isnumber(k) for k in x])


def _ensure_list(x):
    return [x] if not _issequence(x) else x


def _ensure_string_from_expression(expression):
    import vaex.expression
    if expression is None:
        return None
    elif isinstance(expression, bool):
        return expression
    elif isinstance(expression, six.string_types):
        return expression
    elif isinstance(expression, vaex.expression.Expression):
        return expression.expression
    else:
        raise ValueError('%r is not of string or Expression type, but %r' % (expression, type(expression)))


def _ensure_strings_from_expressions(expressions):
    if _issequence(expressions):
        return [_ensure_strings_from_expressions(k) for k in expressions]
    else:
        return _ensure_string_from_expression(expressions)


def _expand(x, dimension, type=tuple):
    if _issequence(x):
        assert len(x) == dimension, "wants to expand %r to dimension %d" % (x, dimension)
        return type(x)
    else:
        return type((x,) * dimension)


def _expand_shape(shape, dimension):
    if isinstance(shape, (tuple, list)):
        assert len(shape) == dimension, "wants to expand shape %r to dimension %d" % (shape, dimension)
        return tuple(shape)
    else:
        return (shape,) * dimension


def _expand_limits(limits, dimension):
    if isinstance(limits, (tuple, list, np.ndarray)) and \
            (isinstance(limits[0], (tuple, list, np.ndarray)) or isinstance(limits[0], str) or limits[0] is None):
        assert len(limits) == dimension, "wants to expand shape %r to dimension %d" % (limits, dimension)
        return tuple(limits)
    else:
        return (limits, ) * dimension


def as_flat_float(a):
    if a.dtype.type == np.float64 and a.strides[0] == 8:
        return a
    else:
        return a.astype(np.float64, copy=True)

def as_flat_array(a, dtype=np.float64):
    if a.dtype.type == dtype and a.strides[0] == 8:
        return a
    else:
        return a.astype(dtype, copy=True)


def is_contiguous(ar):
    return ar.flags['C_CONTIGUOUS']


def as_contiguous(ar):
    return ar if is_contiguous(ar) else ar.copy()


def _split_and_combine_mask(arrays):
    '''Combines all masks from a list of arrays, and logically ors them into a single mask'''
    masks = [np.ma.getmaskarray(block) for block in arrays if np.ma.isMaskedArray(block)]
    arrays = [block.data if np.ma.isMaskedArray(block) else block for block in arrays]
    mask = None
    if masks:
        mask = masks[0].copy()
        for other in masks[1:]:
            mask |= other
    return arrays, mask

def gen_to_list(fn=None, wrapper=list):
    '''A decorator which wraps a function's return value in ``list(...)``.

    Useful when an algorithm can be expressed more cleanly as a generator but
    the function should return an list.

    Example:

    >>> @gen_to_list
    ... def get_lengths(iterable):
    ...     for i in iterable:
    ...         yield len(i)
    >>> get_lengths(["spam", "eggs"])
    [4, 4]
    >>>
    >>> @gen_to_list(wrapper=tuple)
    ... def get_lengths_tuple(iterable):
    ...     for i in iterable:
    ...         yield len(i)
    >>> get_lengths_tuple(["foo", "bar"])
    (3, 3)

    :param fn: generator function to be converted list or touple
    :wrapper (str/tuple) wrapper:
    :return: list or tuple, depending on the wrapper input parameter
    :rtype: list or tuple
    '''
    def listify_return(fn):
        @functools.wraps(fn)
        def listify_helper(*args, **kw):
            return wrapper(fn(*args, **kw))
        return listify_helper
    if fn is None:
        return listify_return
    return listify_return(fn)


def find_type_from_dtype(namespace, prefix, *dtypes, transient=True, support_non_native=True):
    from .array_types import is_string_type
    non_native = False
    postfix = ""
    for i, dtype in enumerate(dtypes):
        if i != 0:
            postfix += "_"
        if dtype == 'string':
            if transient:
                postfix += 'string'
            else:
                postfix += 'string' # view not support atm
        else:
            dtype = dtype.numpy
            type_name = str(dtype)
            if type_name == '>f8':
                type_name = 'float64'
            if dtype.kind == "M":
                type_name = "int64"
            if dtype.kind == "m":
                type_name = "int64"
            postfix += type_name
            # for object there is no non-native version
            if support_non_native and dtype.kind != 'O' and dtype.byteorder not in ["<", "=", "|"]:
                if i != 0:
                    if not non_native:
                        raise TypeError('Mixed endianness')
                non_native = True
            else:
                if non_native:
                    raise TypeError('Mixed endianness')
    if non_native:
        postfix += "_non_native"
    name = prefix + postfix
    if hasattr(namespace, name):
        return getattr(namespace, name)
    else:
        raise ValueError('Could not find a class (%s), seems %s is not supported' % (name, dtype))


def to_native_dtype(dtype):
    if isinstance(dtype, np.dtype) and dtype.byteorder not in "<=|":
        return dtype.newbyteorder()
    else:
        return dtype


def to_native_array(ar):
    if ar.dtype.byteorder not in "<=|":
        return ar.astype(to_native_dtype(ar.dtype))
    else:
        return ar


def extract_central_part(ar):
    return ar[(slice(2,-1), ) * ar.ndim]

def unmask_selection_mask(selection_mask):
    if np.ma.isMaskedArray(selection_mask):
        # if we are doing a selection on a masked array
        selection_mask, mask = selection_mask.data, np.ma.getmaskarray(selection_mask)
        # exclude the masked values
        selection_mask = selection_mask & ~mask
    return selection_mask


def wrap_future_with_promise(future):
    from vaex.promise import Promise
    if isinstance(future, Promise):  # TODO: not so nice, sometimes we pass a promise
        return future
    promise = Promise()

    def callback(future):
        e = future.exception()
        if e:
            promise.reject(e)
        else:
            promise.fulfill(future.result())
    future.add_done_callback(callback)
    return promise


def required_dtype_for_max(N, signed=True):
    if signed:
        dtypes = [np.int8, np.int16, np.int32, np.int64]
    else:
        dtypes = [np.uint8, np.uint16, np.uint32, np.uint64]
    for dtype in dtypes:
        if N <= np.iinfo(dtype).max:
            return np.dtype(dtype)
    else:
        raise ValueError(f"Cannot store a max value of {N} inside an uint64/int64")


def required_dtype_for_range(vmin, vmax, signed=True):
    if signed:
        dtypes = [np.int8, np.int16, np.int32, np.int64]
    else:
        dtypes = [np.uint8, np.uint16, np.uint32, np.uint64]
    for dtype in dtypes:
        if (vmin >= np.iinfo(dtype).min) and (vmax <= np.iinfo(dtype).max):
            return np.dtype(dtype)
    else:
        raise ValueError(f"Cannot store the range {vmin}-{vmax} inside an uint64/int64")


def required_dtype_for_int(value, signed=True):
    if signed or value < 0:
        dtypes = [np.int8, np.int16, np.int32, np.int64]
    else:
        dtypes = [np.uint8, np.uint16, np.uint32, np.uint64]
    for dtype in dtypes:
        if value >= 0:
            if value <= np.iinfo(dtype).max:
                return np.dtype(dtype)
        else:
            if value >= np.iinfo(dtype).min:
                return np.dtype(dtype)
    else:
        raise ValueError(f"Cannot store a value of {value} inside an uint64/int64")


def print_stack_trace(*args, **kwargs):
    import traceback
    print("args: ", args, kwargs)
    traceback.print_stack()


def print_exception_trace(e):
    import traceback
    import sys
    print(''.join(traceback.format_exception(None, e, e.__traceback__)), file=sys.stdout, flush=True)


def format_exception_trace(e):
    import traceback
    return ''.join(traceback.format_exception(None, e, e.__traceback__))


class ProxyModule:
    def __init__(self, name, version, modules=None):
        self.name = name
        self.module = None
        self.version = version
        self.modules = modules if modules else [self.name]

    def _ensure_import(self):
        if self.module is None:
            import importlib
            try:
                for module_name in self.modules:
                    importlib.import_module(module_name)
                # the module object itself needs to be the top module
                top_name = self.name.split(".")[0]
                self.module = importlib.import_module(top_name)
            except Exception as e:
                raise ImportError(f'''Error importing module {self.name}: {e}

Vaex needs an optional dependency '{self.name}' for the feature you are using. To install, use:

$ pip install "{self.name}{self.version}"

Or when using conda:
$ conda install -c conda-forge "{self.name}{self.version}""

        ''') from e

    def __getattr__(self, name):
        self._ensure_import()
        return getattr(self.module, name)


def optional_import(name, version='', modules=None):
    return ProxyModule(name, version=version, modules=modules)


def div_ceil(n, d):
    """Integer divide that sounds up (to an int).

    See https://stackoverflow.com/a/54585138/5397207

    Examples
    >>> div_ceil(6, 2)
    3
    >>> div_ceil(7, 2)
    4
    >>> div_ceil(8, 2)
    4
    >>> div_ceil(9, 2)
    5
    """
    return (n + d - 1) // d


def get_env_memory(key, default=None):
    value = os.environ.get(key, default)
    if value is not None:
        try:
            value = ast.literal_eval(value)
        except:
            pass
        if isinstance(value, str):
            value = dask.utils.parse_bytes(value)
        if not isinstance(value, int):
            raise TypeError(f"Expected env var {key} to be of integer type")
    return value


def get_env_type(type, key, default=None):
    '''Get an env var named key, and cast to type

    >>> import os
    >>> get_env_type(int, 'VAEX_NUM_THREADS_TEST')
    >>> get_env_type(int, 'VAEX_NUM_THREADS_TEST', 10)
    10
    >>> get_env_type(int, 'VAEX_NUM_THREADS_TEST', '10')
    10
    >>> os.environ['VAEX_NUM_THREADS_TEST'] = '20'
    >>> get_env_type(int, 'VAEX_NUM_THREADS_TEST')
    20
    >>> get_env_type(int, 'VAEX_NUM_THREADS_TEST', '10')
    20
    >>> os.environ['VAEX_NUM_THREADS_TEST'] = ' '
    >>> get_env_type(int, 'VAEX_NUM_THREADS_TEST')
    >>> get_env_type(int, 'VAEX_NUM_THREADS_TEST', '11')
    11
    '''
    value = os.environ.get(key, default)
    if isinstance(value, str) and value.strip() == '' and type != str:
        # support empty strings
        value = default
    if value is not None:
        return type(ast.literal_eval(repr(value)))


def dropnan(sequence, expect=None):
    original_type = type(sequence)
    sequence = list(sequence)
    non_nan = [k for k in sequence if k == k]
    if expect is not None:
        assert len(sequence) - len(non_nan) == 1, "expected 1 nan value"
    return original_type(non_nan)


def dict_replace_key(d, key_old, key_new):
    '''Replace a key, without changing order'''
    d_new = {}
    for key, value in d.items():
        if key == key_old:
            key = key_new
        d_new[key] = value
    return d_new

# backwards compatibility
def progressbars(*args, **kwargs):
    from .progress import tree
    return tree(*args, **kwargs)


import contextlib
@contextlib.contextmanager
def file_lock(name):
    '''Context manager for creating a file lock in the file lock directory.

    :param name: A unique name for the context (e.g. a fingerprint) on which the filename is based.
    '''
    path = os.path.join(vaex.settings.main.path_lock, name)
    os.makedirs(os.path.dirname(path), exist_ok=True)
    with FileLock(path):
        yield


def create_hasher(initial_data=None, large_data=True):
    """Creates a blake3 hasher"""
    if not hasattr(blake3, "__version__"):
        version = (0, 2)
    else:
        version = tuple(map(int, blake3.__version__.split(".")[:2]))
    if version < (0, 3):
        return blake3.blake3(initial_data, multithreading=large_data)
    else:
        return blake3.blake3(initial_data, max_threads=blake3.blake3.AUTO if large_data else 1)



from __future__ import division, print_function
import ast
import asyncio
from collections import defaultdict
import contextlib
import os
import time
import threading
import math
import multiprocessing
import logging
import queue
from typing import List

import dask.utils
import numpy as np

import vaex.asyncio
import vaex.cpu  # force registration of task-part-cpu
import vaex.encoding
import vaex.memory
import vaex.multithreading
import vaex.vaexfast
import vaex.events
import vaex.settings

try:
    # support py 36 for the moment
    import contextvars
    has_contextvars = True
except ModuleNotFoundError:
    has_contextvars = False


logger = logging.getLogger("vaex.execution")


class UserAbort(Exception):
    def __init__(self, reason):
        super(UserAbort, self).__init__(reason)


class Run:
    def __init__(self, tasks):
        self.tasks = tasks
        self.cancelled = False
        dataset = self.tasks[0].df.dataset
        if self.tasks[0].df._index_start != 0 or self.tasks[0].df._index_end != dataset.row_count:
            dataset = dataset.slice(self.tasks[0].df._index_start, self.tasks[0].df._index_end)
        for task in tasks:
            assert self.tasks[0].df._index_start == task.df._index_start
            assert self.tasks[0].df._index_end == task.df._index_end
        self.dataset = dataset
        self.pre_filter_per_df = {}
        self.expressions = list(set(expression for task in tasks for expression in task.expressions_all))
        self.tasks_per_df = dict()
        for task in tasks:
            if task.df not in self.tasks_per_df:
                self.tasks_per_df[task.df] = []
            self.tasks_per_df[task.df].append(task)
        for task in tasks:
            if task.df not in self.pre_filter_per_df:
                self.pre_filter_per_df[task.df] = task.pre_filter
            else:
                if self.pre_filter_per_df[task.df] != task.pre_filter:
                    raise ValueError(f"All tasks need to be pre_filter'ed or not pre_filter'ed, it cannot be mixed: {tasks}")


        # find the columns from the dataset we need
        dfs = set()
        for task in tasks:
            dfs.add(task.df)
        # shared between all datasets (TODO: can we detect the filter is cached already?)
        self.dataset_deps = set()
        # per dataframe:
        self.filter_deps = {}
        self.selection_deps = {}
        self.expression_deps = {}

        for df in dfs:
            self.filter_deps[df] = set()
            self.selection_deps[df] = set()
            self.expression_deps[df] = set()
            others = set(df.variables) | set(df.virtual_columns) | set(df.selection_histories)
            tasks_df = [task for task in tasks if task.df == df]
            expressions = list(set(expression for task in tasks_df for expression in task.expressions_all))
            selections = list(set(selection for task in tasks_df for selection in task.selections))
            variables = set()
            for expression in expressions:
                variables |= df._expr(expression).expand().variables(ourself=True)
            for var in variables:
                if var not in self.dataset:
                    if var not in others:
                        raise RuntimeError(f'Oops, requesting column {var} from dataset, but it does not exist')
                    else:
                        pass  # ok, not a column, just a var or virtual column
                else:
                    self.dataset_deps.add(var)
                    self.expression_deps[df].add(var)

            variables = set()
            if df.filtered:
                variables = df.get_selection(vaex.dataframe.FILTER_SELECTION_NAME).dependencies(df)
                for var in variables:
                    if var not in self.dataset:
                        if var not in others:
                            raise RuntimeError(f'Oops, requesting column {var} from dataset, but it does not exist')
                        else:
                            pass  # ok, not a column, just a var or virtual column
                    else:
                        self.dataset_deps.add(var)
                        self.filter_deps[df].add(var)
            variables = set()
            for selection in selections:
                if selection is not None:
                    variables |= df._selection_expression(selection).dependencies()
            for var in variables:
                if var not in self.dataset:
                    if var not in others:
                        raise RuntimeError(f'Oops, requesting column {var} from dataset, but it does not exist')
                    else:
                        pass  # ok, not a column, just a var or virtual column
                else:
                    self.dataset_deps.add(var)
                    self.selection_deps[df].add(var)

        logger.debug('Using columns %r from dataset', self.dataset_deps)


def _merge(tasks):
    dfs = set()
    for task in tasks:
        dfs.add(task.df)
    tasks_merged = []
    for df in dfs:
        tasks_df = [k for k in tasks if k.df == df]
        tasks_merged.extend(_merge_tasks_for_df(tasks_df, df))
    return tasks_merged


def _merge_tasks_for_df(tasks, df):
    # non-mergable:
    tasks_non_mergable = [task for task in tasks if not isinstance(task, vaex.tasks.TaskAggregation)]
    # mergable
    tasks_agg = [task for task in tasks if isinstance(task, vaex.tasks.TaskAggregation)]
    tasks_merged = []

    # Merge aggregation tasks to single aggregations tasks if possible
    tasks_agg_per_grid = defaultdict(list)
    for task in tasks_agg:
        tasks_agg_per_grid[task.binners].append(task)
    for binners, tasks in tasks_agg_per_grid.items():
        task_merged = vaex.tasks.TaskAggregations(df, binners)
        task_merged.original_tasks = tasks
        for i, subtask in enumerate(tasks):
            # chain progress
            def progress_wrapper(p, subtask=subtask):
                result = all(subtask.signal_progress.emit(p))
                return result
            task_merged.signal_progress.connect(progress_wrapper)
            task_merged.add_aggregation_operation(subtask.aggregation_description)
            @vaex.delayed
            def assign(value, i=i, subtask=subtask):
                subtask.fulfill(value[i])
            assign(task_merged)
            task_merged.done(None, subtask.reject)
            task_merged.signal_start.connect(subtask.signal_start.emit)
        tasks_merged.append(task_merged)
    return tasks_non_mergable + tasks_merged


class Executor:
    """An executor is responsible to executing tasks, they are not reentrant, but thread safe"""
    def __init__(self, async_method=None):
        self.tasks : List[Task] = []
        self.async_method = async_method
        self.signal_begin = vaex.events.Signal("begin")
        self.signal_progress = vaex.events.Signal("progress")
        self.signal_end = vaex.events.Signal("end")
        self.signal_cancel = vaex.events.Signal("cancel")
        self.local = threading.local()  # to make it non-reentrant
        if has_contextvars:
            # used for calling execute_async from different async callstacks
            self.isnested = contextvars.ContextVar('executor', default=False)
        self.lock = threading.Lock()
        self.event_loop = asyncio.new_event_loop()

    if hasattr(contextlib, 'asynccontextmanager'):
        @contextlib.asynccontextmanager
        async def auto_execute(self):
            '''This async executor will start executing tasks automatically when a task is awaited for.'''
            vaex.promise.auto_await_executor.set(self)
            try:
                yield
                await self.execute_async()
            finally:
                vaex.promise.auto_await_executor.set(None)
    else:
        async def auto_execute(self):
            raise NotImplemented('Only on Python >3.7')

    async def execute_async(self):
        raise NotImplementedError

    def execute(self):
        raise NotImplementedError

    def run(self, coro):
        async_method = self.async_method or vaex.settings.main.async_
        if async_method == "nest":
            return vaex.asyncio.just_run(coro)
        elif async_method == "awaitio":
            with vaex.asyncio.with_event_loop(self.event_loop):
                return self.event_loop.run_until_complete(coro)
        else:
            raise RuntimeError(f'No async method: {async_default}')

    def schedule(self, task):
        '''Schedules new task for execution, will return an existing tasks if the same task was already added'''
        with self.lock:
            for task_existing in self.tasks:
                if (task_existing.df == task.df) and task_existing.fingerprint() == task.fingerprint():
                    key = task.fingerprint()
                    logger.debug("Did not add already existing task with fingerprint: %r", key)
                    return task_existing

            if vaex.cache.is_on() and task.cacheable:
                key = task.fingerprint()                
                logger.debug("task fingerprint: %r (dataset fp=%r)", key, task.df.dataset.fingerprint)
                logger.debug("task repr: %r", task)
                result = vaex.cache.get(key, type="task")
                if result is not None:
                    logger.info("task not added, used cache key: %r", key)
                    task.fulfill(result)
                    return task
                else:
                    logger.info("task *will* be scheduled with cache key: %r", key)
            if task not in self.tasks:
                self.tasks.append(task)
                logger.info("task added, queue: %r", self.tasks)
                return task

    def _pop_tasks(self):
        # returns a list of tasks that can be executed in 1 pass over the data
        # (which currently means all tasks for 1 dataset) and drop them from the
        # list of tasks
        with self.lock:
            if len(self.tasks) == 0:
                return []
            else:
                dataset = self.tasks[0].df.dataset
                i1, i2 = self.tasks[0].df._index_start, self.tasks[0].df._index_end
                # if we have the same dataset *AND* we want to have the same slice out of it
                # we can share it in in single pass over the dataset
                def shares_dataset(df, i1=i1, i2=i2, dataset=dataset):
                    return dataset == df.dataset and i1 == df._index_start and i2 == df._index_end

                tasks = [task for task in self.tasks if shares_dataset(task.df)]
                logger.info("executing tasks in run: %r", tasks)
                for task in tasks:
                    self.tasks.remove(task)
                return tasks


class ExecutorLocal(Executor):
    def __init__(self, thread_pool=None, chunk_size=None, chunk_size_min=None, chunk_size_max=None, thread_mover=None, zigzag=True):
        super().__init__()
        self.thread_pool = thread_pool or vaex.multithreading.ThreadPoolIndex()
        self.chunk_size = chunk_size
        self.chunk_size_min = chunk_size_min
        self.chunk_size_max = chunk_size_max
        self.thread = None
        self.passes = 0  # how many times we passed over the data
        self.zig = True  # zig or zag
        self.zigzag = zigzag
        self.event_loop = asyncio.new_event_loop()

    def _cancel(self, run):
        logger.debug("cancelling")
        self.signal_cancel.emit()
        run.cancelled = True

    def chunk_size_for(self, row_count):
        chunk_size = self.chunk_size or vaex.settings.main.chunk.size
        chunk_size_min = self.chunk_size or vaex.settings.main.chunk.size_min
        chunk_size_max = self.chunk_size or vaex.settings.main.chunk.size_max
        if chunk_size is None:
            # we determine it automatically by defaulting to having each thread do 1 chunk
            chunk_size_1_pass = vaex.utils.div_ceil(row_count, self.thread_pool.nthreads)
            # brackated by a min and max chunk_size
            chunk_size = min(chunk_size_max, max(chunk_size_min, chunk_size_1_pass))
        return chunk_size

    async def execute_async(self):
        # consume awaitables from the generator, and await them here at the top
        # level, such that we don't need await in the downstream code
        # so we can reuse the same code in a sync way
        gen = self.execute_generator(use_async=True)
        value = None
        while True:
            try:
                value = gen.send(value)
                value = await value
            except StopIteration:
                break

    def execute(self):
        for _ in self.execute_generator():
            pass  # just eat all elements

    def execute_generator(self, use_async=False):
        logger.debug("starting with execute")

        with self.lock:  # setup thread local initial values
            if not hasattr(self.local, 'executing'):
                self.local.executing = False

        try:
            t0 = time.time()
            self.local.cancelled = False
            self.signal_begin.emit()
            # keep getting a list of tasks
            # we currently process tasks (grouped) per df
            # but also, tasks can add new tasks
            while True:
                tasks = self.local.tasks = self._pop_tasks()

                # wo don't allow any thread from our thread pool to enter (a computation should never produce a new task)
                # and we explicitly disallow reentry (this usually means a bug in vaex, or bad usage)
                chunk_executor_thread = threading.current_thread() in self.thread_pool._threads
                import traceback
                trace = ''.join(traceback.format_stack())
                if chunk_executor_thread or self.local.executing and (has_contextvars is False or self.isnested.get() is True):
                    logger.error("nested execute call")
                    raise RuntimeError("nested execute call: %r %r\nlast trace:\n%s\ncurrent trace:\n%s" % (chunk_executor_thread, self.local.executing, self.local.last_trace, trace))
                else:
                    self.local.last_trace = trace

                self.local.executing = True
                if has_contextvars:
                    self.isnested.set(True)

                if not tasks:
                    break
                tasks = _merge(tasks)
                run = Run(tasks)
                self.passes += 1
                dataset = run.dataset

                run.variables = {}
                for df in run.tasks_per_df.keys():
                    run.variables[df] = {key: df.evaluate_variable(key) for key in df.variables.keys()}

                # (re) thrown exceptions as soon as possible to avoid complicated stack traces
                for task in tasks:
                    if task.isRejected:
                        task.get()
                    if hasattr(task, "check"):
                        try:
                            task.check()
                        except Exception as e:
                            task.reject(e)
                            raise

                for task in run.tasks:
                    task.signal_start.emit(self)

                for task in run.tasks:
                    task._results = []
                    if not any(task.signal_progress.emit(0)):
                        logger.debug("task cancelled immediately")
                        task.cancelled = True
                row_count = dataset.row_count
                chunk_size = self.chunk_size_for(row_count)
                encoding = vaex.encoding.Encoding()
                run.nthreads = nthreads = self.thread_pool.nthreads
                task_checkers = vaex.tasks.create_checkers()
                memory_tracker = vaex.memory.create_tracker()
                vaex.memory.local.agg = memory_tracker
                # we track this for consistency
                memory_usage = 0
                for task in tasks:
                    for task_checker in task_checkers:
                        task_checker.add_task(task)
                    spec = encoding.encode('task', task)
                    spec['task-part-cpu-type'] = spec.pop('task-type')
                    def create_task_part():
                        nonlocal memory_usage
                        task_part = encoding.decode('task-part-cpu', spec, df=task.df, nthreads=nthreads)
                        memory_usage += task_part.memory_usage()
                        for task_checker in task_checkers:
                            task_checker.add_task(task)
                        if task.requires_fingerprint:
                            task_part.fingerprint = task.fingerprint()
                        return task_part
                    # We want at least 1 task part (otherwise we cannot do any work)
                    # then we ask for the task part how often we should split
                    # This means that we can have 100 threads, but only 2 task parts
                    # In this case, evaluation of expressions is still multithreaded,
                    # but aggregation is reduced to effectively 2 threads.
                    task_part_0 = create_task_part()
                    ideal_task_splits = task_part_0.ideal_splits(self.thread_pool.nthreads)
                    assert ideal_task_splits <= self.thread_pool.nthreads, f'Cannot have more splits {ideal_task_splits} then threads {self.thread_pool.nthreads}'
                    if ideal_task_splits == self.thread_pool.nthreads or task.see_all:
                        # in the simple case, we just use a list
                        task._parts = [task_part_0] + [create_task_part() for i in range(1, ideal_task_splits)]
                    else:
                        # otherwise a queue
                        task._parts = queue.Queue()
                        task._parts.put(task_part_0)
                        for i in range(1, ideal_task_splits):
                            task._parts.put(create_task_part())
                if memory_usage != memory_tracker.used:
                    raise RuntimeError(f"Reported memory usage by tasks was {memory_usage}, while tracker listed {memory_tracker.used}")
                vaex.memory.local.agg = None

                # TODO: in the future we might want to enable the zigzagging again, but this requires all datasets to implement it
                # if self.zigzag:
                #     self.zig = not self.zig
                def progress(p):
                    # no global cancel and at least 1 tasks wants to continue, then we continue
                    ok_tasks = any([task.progress(p) for task in tasks])
                    all_stopped = all([task.stopped for task in tasks])
                    ok_executor = all(self.signal_progress.emit(p))
                    if all_stopped:
                        logger.debug("Pass cancelled because all tasks are stopped: %r", tasks)
                    if not ok_tasks:
                        logger.debug("Pass cancelled because all tasks cancelled: %r", tasks)
                    if not ok_executor:
                        logger.debug("Pass cancelled because of the global progress event: %r", self.signal_progress.callbacks)
                    return ok_tasks and ok_executor and not all_stopped
                yield from self.thread_pool.map(self.process_part, dataset.chunk_iterator(run.dataset_deps, chunk_size),
                                                    dataset.row_count,
                                                    progress=progress,
                                                    cancel=lambda: self._cancel(run), unpack=True, run=run, use_async=use_async)
                duration_wallclock = time.time() - t0
                logger.debug("executing took %r seconds", duration_wallclock)
                self.local.executing = False
                if has_contextvars:
                    self.isnested.set(False)
                if True:  # kept to keep the diff small
                    for task in tasks:
                        if not task.cancelled:
                            logger.debug("fulfill task: %r", task)
                            parts = task._parts
                            if not isinstance(parts, list):
                                parts_queue = parts
                                parts = []
                                while not parts_queue.empty():
                                    parts.append(parts_queue.get())
                            parts[0].reduce(parts[1:])
                            logger.debug("wait for task: %r", task)
                            task._result = parts[0].get_result()
                            task.end()
                            task.fulfill(task._result)
                            logger.debug("got result for: %r", task)
                            if task._result is not None and task.cacheable:  # we don't want to store None
                                if vaex.cache.is_on():
                                    # we only want to store the original task results into the cache
                                    tasks_cachable = task.original_tasks if isinstance(task, vaex.tasks.TaskAggregations) else [task]
                                    for task_cachable in tasks_cachable:
                                        key = task_cachable.fingerprint()
                                        previous_result = vaex.cache.get(key, type='task')
                                        if (previous_result is not None):# and (previous_result != task_cachable.get()):
                                            try:
                                                if previous_result != task_cachable.get():
                                                    # this can happen with multithreading, where two threads enter the same tasks in parallel (IF using different executors)
                                                    logger.warning("calculated new result: %r, while cache had value: %r", previous_result, task_cachable.get())
                                            except ValueError:  # when comparing numpy results
                                                if not np.array_equal(previous_result, task_cachable.get(), equal_nan=True):
                                                    # this can happen with multithreading, where two threads enter the same tasks in parallel (IF using different executors)
                                                    logger.warning("calculated new result: %r, while cache had value: %r", previous_result, task_cachable.get())
                                        vaex.cache.set(key, task_cachable.get(), type='task', duration_wallclock=duration_wallclock)
                                        logger.info("added result: %r in cache under key: %r", task_cachable.get(), key)

                        else:
                            logger.debug("rejecting task: %r", task)
                            # we now reject, in the main thread
                            if task._toreject:
                                task.reject(task._toreject)
                            else:
                                task.reject(UserAbort("Task was cancelled"))
                            # remove references
                        task._result = None
                        task._results = None
                    self.signal_end.emit()
        except:  # noqa
            self.signal_cancel.emit()
            raise
        finally:
            self.local.executing = False
            if has_contextvars:
                self.isnested.set(False)

    def process_part(self, thread_index, i1, i2, chunks, run):
        if not run.cancelled:
            if thread_index >= run.nthreads:
                raise ValueError(f'thread_index={thread_index} while only having {run.nthreads} blocks')
            for df, tasks in run.tasks_per_df.items():
                self.process_tasks(thread_index, i1, i2, chunks, run, df, tasks)
        return i2 - i1

    def process_tasks(self, thread_index, i1, i2, chunks, run, df, tasks):
        if 1:  # avoid large diff
            if i1 == i2:
                raise RuntimeError(f'Oops, get an empty chunk, from {i1} to {i2}, that should not happen')
            N = i2 - i1
            for name, chunk in chunks.items():
                assert len(chunk) == N, f'Oops, got a chunk ({name}) of length {len(chunk)} while it is expected to be of length {N} (at {i1}-{i2}'

            assert df in run.variables
            from .scopes import _BlockScope

            expressions = list(set(expression for task in tasks for expression in task.expressions_all))
            pre_filter = tasks[0].pre_filter
            if pre_filter:
                filter_deps = run.filter_deps[df]
                filter_scope = _BlockScope(df, i1, i2, None, selection=True, values={**run.variables[df], **{k: chunks[k] for k in filter_deps if k in chunks}})
                filter_scope.filter_mask = None
                filter_mask = filter_scope.evaluate(vaex.dataframe.FILTER_SELECTION_NAME)
                deps = run.expression_deps[df] | run.selection_deps[df]
                chunks = {k:vaex.array_types.filter(v, filter_mask) for k, v, in chunks.items() if k in deps}
            else:
                filter_mask = None
            def sanity_check(name, chunk):
                if not vaex.column.is_column_like(chunk):
                    raise TypeError(f'Evaluated a chunk ({name}) that is not an array of column like object: {chunk!r} (type={type(chunk)}')
            for name, chunk in chunks.items():
                sanity_check(name, chunk)
            block_scope = _BlockScope(df, i1, i2, values={**run.variables[df], **chunks})
            block_scope.mask = filter_mask
            block_dict = {expression: block_scope.evaluate(expression) for expression in expressions}
            selection_scope = _BlockScope(df, i1, i2, None, selection=True, values={**block_scope.values})
            selection_scope.filter_mask = filter_mask
            if not pre_filter and df.filtered:
                filter_mask = selection_scope.evaluate(vaex.dataframe.FILTER_SELECTION_NAME)

            memory_tracker = vaex.memory.create_tracker()
            task_checkers = vaex.tasks.create_checkers()
            for task in tasks:
                if task.stopped:
                    continue
                assert df is task.df
                blocks = [block_dict[expression] for expression in task.expressions_all]
                def fix(ar):
                    if np.ma.isMaskedArray(ar):
                        ar = vaex.utils.unmask_selection_mask(ar)
                    return ar
                selections = [None if s is None else fix(selection_scope.evaluate(s)) for s in task.selections]
                if not run.cancelled:
                    if task.see_all:
                        assert isinstance(task._parts, list)
                        task_parts = task._parts
                    else:
                        if isinstance(task._parts, list):
                            task_parts = [task._parts[thread_index]]
                        else:
                            task_parts = [task._parts.get()]
                    some_parts_stopped = False
                    for task_index, task_part in enumerate(task_parts):
                        if not task_part.stopped:
                            try:
                                if task.see_all:
                                    task_part.process(thread_index, i1, i2, filter_mask, selections, blocks)
                                else:
                                    task_part.process(thread_index, i1, i2, filter_mask, selections, blocks)
                            except Exception as e:
                                # we cannot call .reject, since then we'll handle fallbacks in this thread
                                task._toreject = e
                                task.cancelled = True
                            finally:
                                if not isinstance(task._parts, list):
                                    task._parts.put(task_part)
                            # we could be done after processing
                            if task_part.stopped:
                                some_parts_stopped = True
                        else:
                            some_parts_stopped = True
                        if some_parts_stopped:
                            break
                    if some_parts_stopped:  # if 1 is done, the whole task is done
                        task.stopped = True
                if memory_tracker.track_live:
                    for part in task_parts:
                        memory_tracker.using(part.memory_usage())
                for task_checker in task_checkers:
                    if task_checker.track_live:
                        for part in task_parts:
                            task_checker.add_task_part(part)



import pyarrow as pa
from .utils import list_unwrap

def test_list_wrap_basics():
    data = ['aap', 'noot', None], None, [], ['aap', 'noot', 'mies']
    string_list = pa.array(data)
    values, wrapper = list_unwrap(string_list)
    assert values.tolist() == string_list.flatten().tolist()

    data_reversed = [['paa', 'toon', None], None, [], ['paa', 'toon', 'seim']]
    reversed = [k[::-1] if k else None for k in values.tolist()]
    assert wrapper(pa.array(reversed)).tolist() == data_reversed


def test_list_wrap_slice():
    data = ['aap', 'noot', None], None, [], ['aap', 'noot', 'mies']
    string_list = pa.array(data)
    values, wrapper = list_unwrap(string_list.slice(1))
    assert values.tolist() == string_list.slice(1).flatten().tolist()

    data_reversed = [None, [], ['paa', 'toon', 'seim']]
    reversed = [k[::-1] if k else None for k in values.tolist()]
    assert wrapper(pa.array(reversed)).tolist() == data_reversed



def test_list_list_wrap():
    data = [['aap', 'noot', None], None, [], ['aap', 'noot', 'mies']], [], None
    string_list_list = pa.array(data)
    values, wrapper = list_unwrap(string_list_list)
    assert values.tolist() == string_list_list.flatten().flatten().tolist()    
    assert wrapper(values).tolist() == string_list_list.tolist()




__author__ = 'maartenbreddels'
import os
import mmap
import collections
import logging
import numpy as np
import numpy.ma
import threading
import vaex
import vaex.dataset
import vaex.file
from vaex.expression import Expression
from vaex.file.column import ColumnFile
import struct

from vaex.multithreading import get_main_io_pool
from vaex.itertools import pmap, buffer, pwait

logger = logging.getLogger("vaex.file")

dataset_type_map = {}

osname = vaex.utils.osname
no_mmap = not vaex.settings.main.mmap



class DatasetMemoryMapped(vaex.dataset.DatasetFile):
    """Represents a dataset where the data is memory mapped for efficient reading"""

    def __init__(self, path, write=False, nommap=False, fs_options={}, fs=None):
        super().__init__(path=path, write=write, fs_options=fs_options, fs=fs)
        self.nommap = nommap
        self.file_map = {}
        self.fileno_map = {}
        self.mapping_map = {}
        self.tls_map = collections.defaultdict(threading.local)

    def chunk_iterator(self, columns, chunk_size=None, reverse=False):
        if self.nommap:
            # we expect here a path like s3 fetching, which will benefit from multithreading
            pool = get_main_io_pool()
            def read(i1, i2, reader):
                return i1, i2, reader()
            chunks_generator = self._default_lazy_chunk_iterator(self._columns, columns, chunk_size)
            yield from pwait(buffer(pmap(read, chunks_generator, pool), pool._max_workers+3))
        else:
            yield from self._default_chunk_iterator(self._columns, columns, chunk_size, reverse=reverse)

    def __getstate__(self):
        return {
            **super().__getstate__(),
            'nommap': self.nommap
        }

    def __setstate__(self, state):
        super().__setstate__(state)
        self.mapping_map = {}
        self.fileno_map = {}
        self.file_map = {}


    def _get_file(self, path):
        assert self.nommap
        if path not in self.file_map:
            file = open(path, "rb+" if self.write else "rb")
            self.file_map[path] = file
        return self.file_map[path]

    def _get_mapping(self, path):
        assert not self.nommap
        if path not in self.mapping_map:
            file = open(path, "rb+" if self.write else "rb")
            fileno = file.fileno()
            kwargs = {}
            if vaex.utils.osname == "windows":
                kwargs["access"] = mmap.ACCESS_READ | 0 if not self.write else mmap.ACCESS_WRITE
            else:
                kwargs["prot"] = mmap.PROT_READ | 0 if not self.write else mmap.PROT_WRITE
            mapping = mmap.mmap(fileno, 0, **kwargs)
            # TODO: we can think about adding this in py38
            # mapping.madvise(mmap.MADV_SEQUENTIAL)
            self.file_map[path] = file
            self.fileno_map[path] = fileno
            self.mapping_map[path] = mapping
        return self.mapping_map[path]

    def close(self):
        for name, memmap in self.mapping_map.items():
            memmap.close()
        for name, file in self.file_map.items():
            file.close()

    def _map_array(self, offset=None, shape=None, dtype=np.float64, path=None):
        if path is None:
            path = self.path
        return self._do_map(path, offset, shape, dtype)
    
    def _do_map(self, path, offset, shape, dtype):
        length = np.product(shape)
        if self.nommap:
            if len(shape) > 1:
                raise RuntimeError('not supported, high d arrays from non local files')
            file = self._get_file(path)
            column = ColumnFile(file, offset, length, dtype, write=self.write, path=self.path, tls=self.tls_map[path])
        else:
            mapping = self._get_mapping(path)
            column = np.frombuffer(mapping, dtype=dtype, count=length, offset=offset)
            column = column.reshape(shape)
        return column



from typing import Dict, List

import pyarrow as pa
import numpy as np
from frozendict import frozendict

import vaex.dataset


def _check(i1, i2):
    if i1 == i2:
        raise RuntimeError(f'Oops, get an empty chunk, from {i1} to {i2}, that should not happen')

def chunk_copy(chunk_iter, mapping):
    for i1, i2, chunks in chunk_iter:
        _check(i1, i2)
        chunks = dict(chunks)
        chunks.update({mapping[name]: chunks[name] for name in mapping if name in chunks})
        yield i1, i2, chunks


def chunk_project(chunk_iter, keys):
    for i1, i2, chunks in chunk_iter:
        _check(i1, i2)
        yield i1, i2, {name: chunks[name] for name in keys}


def chunk_trim(chunk_iter, count):
    seen = 0
    for i1, i2, chunks in chunk_iter:
        _check(i1, i2)
        if i2 > count:  # last chunk
            max_length = count - i1
            if max_length == 0:
                return
            chunks = {name: vaex.array_types.slice(chunks[name], 0, max_length) for name in chunks}
            i2 = i1 + max_length
            _check(i1, i2)
            yield i1, i2, chunks
            return
        else:
            _check(i1, i2)
            yield i1, i2, chunks


def chunk_eat(chunk_iter, count):
    trimmed = 0
    for i1, i2, chunks in chunk_iter:
        _check(i1, i2)
        chunk_length = len(list(chunks.values())[0])
        # do we still have to trim stuff off?
        if trimmed < count:
            to_trim = count - trimmed
            # if so, can we skip this chunk?
            if to_trim >= chunk_length:
                trimmed += chunk_length
                continue
            else:
                # otherwise we just trim off the begin of this chunk
                chunks = {name: vaex.array_types.slice(chunks[name], to_trim) for name in chunks}
                trimmed += to_trim
                i1 += to_trim
        _check(i1 - count, i2 - count)
        yield i1 - count, i2 - count, chunks


def sliding_matrix(ar_prev, ar, ar_next, steps, offset, fill_value=None):
    ar = vaex.array_types.to_numpy(ar)
    dtype = vaex.dtype_of(ar)
    assert ar.ndim == 1
    N = len(ar)
    if fill_value is None:
        M = np.empty(ar.shape + (steps,), dtype=dtype.numpy)
        M = np.ma.array(M, shrink=False, mask=np.full(M.shape, True))
    else:
        M = np.full(ar.shape + (steps,), fill_value, dtype=dtype.numpy)

    if ar_next is not None:
        ar_next = vaex.array_types.to_numpy(ar_next)

    for i in range(steps):
        if ar_prev is not None:
            part_prev = ar_prev[-offset+i:]
            M[:len(part_prev),i] = part_prev
        part = ar[max(0, i-offset):len(ar)-offset+i]
        start = max(offset-i, 0)
        M[start:start+len(part),i] = part
        if ar_next is not None:
            start += len(part)
            part_next = ar_next[:len(M)-start]
            M[start:start+len(part_next),i] = part_next

    return M

def chunk_sliding_matrix(chunk_iter, shift, keys, chunk_size, offset=0, fill_value=None):
    assert shift < chunk_size
    chunks_ready_list_prepend = []
    i1, i2, chunks = next(chunk_iter)
    chunks_prev = {}
    for i1_next, i2_next, chunks_next in chunk_iter:
        _check(i1_next, i2_next)
        chunks_extra = {name: sliding_matrix(chunks_prev.get(name), chunks[name], chunks_next[name], shift, offset=offset, fill_value=fill_value) for name in keys}
        _check(i1, i2)
        yield i1, i2, {**chunks, **chunks_extra}
        chunks_prev = chunks
        chunks = chunks_next
        i1 = i1_next
        i2 = i2_next
    chunks_extra = {name: sliding_matrix(None, chunks[name], None, shift, offset=offset, fill_value=fill_value) for name in keys}
    _check(i1, i2)
    yield i1, i2, {**chunks, **chunks_extra}


def chunk_prepend(chunk_iter, prepend_chunks, chunk_size, trim=True):
    chunks_ready_list_prepend = []
    chunks_ready_list_prepend.append(prepend_chunks)
    for i1, i2, chunks in chunk_iter:
        chunks = dict(chunks)
        for chunk in chunks.values():
            assert len(chunk) <= chunk_size, f'Expected chunk_size<={chunk_size}, but got {len(chunk)}'
        chunks_ready_list_prepend.append({name: ar for name, ar in chunks.items() if name in prepend_chunks})
        chunks_passthrough = {name: ar for name, ar in chunks.items() if name not in prepend_chunks}
        chunks_current_list_prepend, current_row_count_prepend = vaex.dataset._slice_of_chunks(chunks_ready_list_prepend, chunk_size)
        chunks_prepend = vaex.dataset._concat_chunk_list(chunks_current_list_prepend)
        chunks = {**chunks_prepend, **chunks_passthrough}
        length = i2 - i1
        def trim(ar):
            if len(ar) > length:
                return vaex.array_types.slice(ar, 0, length)
            else:
                return ar
        # and trim off the trailing part
        chunks = {name: trim(chunks[name]) for name in chunks}  # sort and trim
        yield i1, i2, chunks


def chunk_append(chunk_iter, append_chunks, chunk_size):
    trimmed = 0
    n = len(list(append_chunks.values())[0])

    chunks_ready_list_append = []
    chunks_ready_list_passthrough = []
    our_i1, our_i2 = 0, 0
    for i1, i2, chunks in chunk_iter:
        chunks = dict(chunks)

        # keep a list of chunks for which we want to append to
        chunks_ready_list_append.append({name: ar for name, ar in chunks.items() if name in append_chunks})
        # and passthrough
        chunks_ready_list_passthrough.append({name: ar for name, ar in chunks.items() if name not in append_chunks})

        if trimmed < n:
            # eat n rows
            _, trimmed_now = vaex.dataset._slice_of_chunks(chunks_ready_list_append, n)
            if trimmed_now == 0:
                # continue getting chunks until we have enough
                # continue  # TODO: how to avoid inf loop
                continue
            else:
                # otherwise we are done
                trimmed = n
            continue

        chunks_current_list_append, current_row_count_append = vaex.dataset._slice_of_chunks(chunks_ready_list_append, chunk_size)
        chunks_append = vaex.dataset._concat_chunk_list(chunks_current_list_append)
        

        has_passthrough = any(name not in append_chunks for name in chunks)
        if has_passthrough:
            chunks_current_list_passthrough, current_row_count_passthrough = vaex.dataset._slice_of_chunks(chunks_ready_list_passthrough, current_row_count_append)
            chunks_passthrough = vaex.dataset._concat_chunk_list(chunks_current_list_passthrough)
            chunks = {**chunks_append, **chunks_passthrough}
            assert current_row_count_passthrough == current_row_count_append
        else:
            chunks = chunks_append
        our_i2 = our_i1 + current_row_count_append
        chunks = {name: chunks[name] for name in chunks}  # sort
        _check(our_i1, our_i2)
        yield our_i1, our_i2, chunks
        our_i1 = our_i2

    # add the trailing part, and repeat the end of the loop
    chunks_ready_list_append.append(append_chunks)
    while chunks_ready_list_append:
        chunks_current_list_append, current_row_count_append = vaex.dataset._slice_of_chunks(chunks_ready_list_append, chunk_size)
        chunks_append = vaex.dataset._concat_chunk_list(chunks_current_list_append)
        

        has_passthrough = any(name not in append_chunks for name in chunks)
        if has_passthrough:
            chunks_current_list_passthrough, current_row_count_passthrough = vaex.dataset._slice_of_chunks(chunks_ready_list_passthrough, chunk_size)
            chunks_passthrough = vaex.dataset._concat_chunk_list(chunks_current_list_passthrough)
            chunks = {**chunks_append, **chunks_passthrough}
            assert current_row_count_passthrough == current_row_count_append
        else:
            chunks = chunks_append
        our_i2 = our_i1 + current_row_count_append
        chunks = {name: chunks[name] for name in chunks}  # sort
        _check(our_i1, our_i2)
        yield our_i1, our_i2, chunks
        our_i1 = our_i2

@vaex.dataset.register
class DatasetShifted(vaex.dataset.DatasetDecorator):
    snake_name = "shift"
    column_mapping: Dict[str, str]  # original name: shifted name
    def __init__(self, original, column_mapping : Dict[str, str], start, end, fill_value=None):
        super().__init__(original)
        self.start = start
        self.end = end
        assert self.start <= self.end
        self.fill_value = fill_value
        self.column_mapping = dict(column_mapping)
        self.column_mapping_reverse = {v: k for k, v in column_mapping.items()}
        self._shifted_column_names = list(column_mapping.values())
        self._ids = {name: ar for name, ar in self.original._ids.items()}
        for original_name, shifted_name in column_mapping.items():
            if original_name in self.original._ids:
                self._ids[shifted_name] = self.original._ids[original_name]
        self._ids = frozendict(self._ids)
        self._create_columns()
        self._row_count = self.original.row_count

    def leafs(self) -> List[vaex.dataset.Dataset]:
        return [self]

    @property
    def _fingerprint(self):
        id = vaex.cache.fingerprint(self.original.id, self.column_mapping, self.start, self.end, self.fill_value)
        return f'dataset-{self.snake_name}-{id}'

    def _encode(self, encoding):
        dataset_spec = encoding.encode('dataset', self.original)
        spec = {'column_mapping': dict(self.column_mapping),
                'fill_value': self.fill_value, 'start': self.start, 'end': self.end,
                'dataset': dataset_spec}
        return spec

    @classmethod
    def _decode(cls, encoding, spec):
        spec = dict(spec)
        dataset = encoding.decode('dataset', spec.pop('dataset'))
        ds = cls(dataset, **spec)
        return ds

    def _create_columns(self):
        for column in self.column_mapping:
            assert column in self.original, f'Expected {column} to be in the current dataset'
        self._columns = {name: vaex.dataset.ColumnProxy(self, name, vaex.array_types.data_type(col)) for name, col in self.original._columns.items()}
        schema = self.original.schema()
        for original_name, shifted_name in self.column_mapping.items():
            self._columns[shifted_name] = vaex.dataset.ColumnProxy(self, shifted_name, schema[original_name])

    def chunk_iterator(self, columns, chunk_size=None, reverse=False, start=0, end=None):
        yield from self._chunk_iterator(columns, chunk_size, start=start, end=end)

    def _chunk_iterator(self, columns, chunk_size, reverse=False, start=0, end=None):
        if start > self.row_count:
            raise ValueError(f'start={start} is >= row_count={self.row_count}')
        end = self.row_count if end is None else end
        chunk_size = chunk_size or 1024**2  # TODO: should we have a constant somewhere
        # if self.n > chunk_size:
            # raise ValueError(f'Iterating with a chunk_size ({chunk_size}) smaller then window_size ({self.window_size})')
        assert not reverse
        # find all columns we need from the original dataset
        columns_original = set(self.column_mapping_reverse.get(name, name) for name in columns)
        # and the once we need to return shifted
        columns_shifted = set(columns) & set(self._shifted_column_names)
        iter = self.original.chunk_iterator(list(columns_original), chunk_size, reverse=reverse)

        # schema for shifted columns only
        schema = {name: vaex.array_types.to_arrow_type(dtype) for name, dtype in self.schema().items()}

        if columns_shifted:
            # (shallow) copy e.g. x to x_shifted
            iter = chunk_copy(iter, self.column_mapping)

            if self.end == self.start:
                if self.start > 0:
                    # for positive shift, we add the fill values at the front
                    iter = chunk_prepend(iter, {name: self._filler(min(self.row_count, self.start), dtype=schema[name]) for name in columns_shifted}, chunk_size)
                elif self.start < 0:
                    # negative at the end
                    # import pdb; pdb.set_trace()
                    iter = chunk_append(iter, {name: self._filler(min(self.row_count, -self.start), dtype=schema[name]) for name in columns_shifted}, chunk_size)
                # for start 0 we don't need to do anything
            else:
                # if self.end > self.start:
                shift = self.end - self.start
                if self.start == 0:
                    iter = chunk_sliding_matrix(iter, shift, columns_shifted, chunk_size, fill_value=self.fill_value)
                elif self.start < 0:
                    # iter = chunk_prepend(iter, {name: self._filler(min(self.row_count, -self.start), dtype=schema[name]) for name in columns_shifted}, chunk_size)
                    # raise "dsa"
                    iter = chunk_sliding_matrix(iter, shift, columns_shifted, chunk_size, offset=-self.start-1, fill_value=self.fill_value)
                else:
                    raise "dsa"
                # das

            # because of the shifting, we may has asked our child dataset for too many columns
            # so we remove them
            iter = chunk_project(iter, columns)
        if start != 0 or end != self.row_count:
            if start != 0:                
                iter = chunk_eat(iter, start)
            if end != self.row_count:
                iter = chunk_trim(iter, end - start)
            iter = vaex.dataset.chunk_rechunk(iter, chunk_size)
        yield from iter

    def _filler(self, n, dtype):
        assert n > 0
        if self.fill_value is None:
            type = vaex.array_types.to_arrow_type(dtype)
            return pa.nulls(n, type=type)
        else:
            return vaex.array_types.full(n, self.fill_value, dtype=dtype)

    def is_masked(self, column):
        column = self.column_mapping_reverse.get(column, column)
        return self.original.is_masked(column) or self.start != self.end

    def shape(self, column):
        column = self.column_mapping_reverse.get(column, column)
        return self.original.shape(column)

    def close(self):
        self.original.close()

    def slice(self, start, end):
        if start == 0 and end == self.row_count:
            return self
        # TODO: we can be smarter here, and trim off some datasets
        return vaex.dataset.DatasetSliced(self, start=start, end=end)

    def hashed(self):
        if set(self._ids) == set(self):
            return self
        return type(self)(self.original.hashed(), column_mapping=self.column_mapping, start=self.start, end=self.end, fill_value=self.fill_value)



import vaex
import pyarrow as pa
import pyarrow.compute as pc


def list_unwrap(ar, level=-1):
    '''Returns the values in a (nested) list, and a callable that puts it back in the same structure'''
    from .convert import trim_offsets
    list_parameters = []
    dtype = vaex.dtype_of(ar)
    array_levels = [ar]
    while dtype.is_list:
        list_parameters.append([ar.type, len(ar), ar.buffers(), ar.null_count, ar.offset])
        # flattened.append(ar.type)
        i1 = ar.offsets[0].as_py()
        i2 = ar.offsets[-1].as_py()
        ar = ar.values.slice(i1, i2)
        array_levels.append(ar)
        dtype = dtype.value_type

    if level == -1:
        ar = array_levels[-1]
    else:
        ar = array_levels[level]
        list_parameters = list_parameters[:level+1]

    def wrapper(new_values):
        if list_parameters and list_parameters:
            ar = None
            for type, length, buffers, null_count, offset in list_parameters[::-1]:
                if ar is None:
                    # buffers = 
                    # assert offset == 0
                    buffers = buffers[:2]
                    buffers = trim_offsets(offset, length, *buffers)
                    offset = 0
                    new_values = vaex.array_types.to_arrow(new_values)
                    type = pa.list_(new_values.type)
                    ar = pa.ListArray.from_buffers(type, length, [buffers[0], buffers[1]], null_count, offset, children=[new_values])
                else:
                    ar = pa.ListArray.from_buffers(type, length, [buffers[0], buffers[1]], null_count, offset, children=[ar])
            return ar
        else:
            return new_values
    return ar, wrapper


def combine_missing(a, b):
    # return a copy of a with missing values of a and b combined
    if a.null_count > 0 or b.null_count > 0:
        a, b = vaex.arrow.convert.align(a, b)
        if isinstance(a, pa.ChunkedArray):
            # divide and conquer
            assert isinstance(b, pa.ChunkedArray)
            assert len(a.chunks) == len(b.chunks)
            return pa.chunked_array([combine_missing(ca, cb) for ca, cb in zip(a.chunks, b.chunks)])
        if a.offset != 0:
            a = vaex.arrow.convert.trim_buffers_ipc(a)
        if b.offset != 0:
            b = vaex.arrow.convert.trim_buffers_ipc(b)
        assert a.offset == 0
        assert b.offset == 0
        # not optimal
        nulls = pc.invert(pc.or_(a.is_null(), b.is_null()))
        assert nulls.offset == 0
        nulls_buffer = nulls.buffers()[1]
        # this is not the case: no reason why it should be (TODO: open arrow issue)
        # assert nulls.buffers()[0] is None
        buffers = a.buffers()
        return pa.Array.from_buffers(a.type, len(a), [nulls_buffer] + buffers[1:])
    else:
        return a
